// Sample4.cpp  Copyright (c) 2005 Kosei Demura 
// My web site is http://demura.net
// This program uses the Open Dynamics Engine (ODE) by Russell Smith.
// The ODE web site is http://ode.org/

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

static dWorldID world;
static dSpaceID space; 
static dGeomID  ground;
static dJointGroupID contactgroup; 
dsFunctions fn;

typedef struct {    
  dBodyID body;     
  dGeomID geom;     
  dReal   radius;
  dReal   length;
  dReal   mass;
} myLink;
myLink torso,leg[2];

dJointID joint[2];

enum jointNo {
  HIP,
  KNEE
};

int time = 0;

static void nearCallback(void *data, dGeomID o1, dGeomID o2)
{
  const int N = 10;
  dContact contact[N];

  int isGround = ((ground == o1) || (ground == o2));

  int n =  dCollide(o1,o2,N,&contact[0].geom,sizeof(dContact));
  if (isGround)  {
    for (int i = 0; i < n; i++) {
      contact[i].surface.mode = dContactBounce;
      contact[i].surface.bounce = 0.8; // (0.0~1.0)
      dJointID c = dJointCreateContact(world,contactgroup,&contact[i]);
      dJointAttach (c,dGeomGetBody(contact[i].geom.g1),
		    dGeomGetBody(contact[i].geom.g2));
    }
  }
}

void control()
{
  double velCoeff = 1.0,p;
  static double angle = 0.5 * M_PI;

  p =  dJointGetHingeAngle(joint[HIP]);
  dReal z = angle  -  p;
  if (p > 0.24 * M_PI) angle = - 0.25 * M_PI;
  else {
    if (p < - 0.24 * M_PI) angle = 0.25 * M_PI;
  }
  dJointSetHingeParam(joint[HIP], dParamVel,  velCoeff*z); 
  dJointSetHingeParam(joint[HIP], dParamFMax, 100);
}

static void simLoop (int pause)
{
  const dReal *pos1,*R1,*pos2,*R2;
  
  time++;

  control(); 

  dSpaceCollide(space,0,&nearCallback);
  dWorldStep(world,0.01);
  dJointGroupEmpty(contactgroup); 
  dsSetColor(1.0,0.0,0.0);

  // draw a torso
  pos1 = dBodyGetPosition(torso.body);
  R1   = dBodyGetRotation(torso.body);
  dsDrawSphereD(pos1,R1,torso.radius);

  // draw legs
  for (int i = 0; i< 2; i++) { 
    pos2 = dBodyGetPosition(leg[i].body);
    R2   = dBodyGetRotation(leg[i].body);
    dsDrawCappedCylinderD(pos2,R2,leg[i].length,leg[i].radius); 
  }
}

void start()
{
  static float xyz[3] = {0.0,-3.0,1.0};
  static float hpr[3] = {90.0,0.0,0.0};
  dsSetViewpoint (xyz,hpr);
}

void  prepDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = NULL;
  fn.stop    = NULL;
  fn.path_to_textures = "../../drawstuff/textures";
}

// Create a monobot
void createMonoBot() {
  dMass m1;
  dReal x0 = 0.0, y0 = 0.0, z0 = 2.5;

  // torso
  torso.radius = 0.2;
  torso.mass   = 10.0;
  torso.body = dBodyCreate(world);   
  dMassSetZero(&m1);
  dMassSetSphereTotal(&m1,torso.mass,torso.radius);
  dBodySetMass(torso.body,&m1);            
  dBodySetPosition(torso.body, x0, y0, z0); 

  torso.geom = dCreateSphere(space,torso.radius); 
  dGeomSetBody(torso.geom,torso.body);       

  // legs
  for (int i = 0; i < 2; i++) {
    leg[i].radius = 0.025;
    leg[i].length = 0.5;
    leg[i].mass   = 1.0;
    leg[i].body = dBodyCreate(world);   
    dMassSetZero(&m1);
    dMassSetCylinderTotal(&m1,leg[i].mass,3,leg[i].radius,leg[i].length);
    dBodySetMass(leg[i].body,&m1);            
    dBodySetPosition(leg[i].body, x0, y0, z0 -torso.radius 
		     - 0.5 * (2 * i + 1) *leg[i].length); 
    leg[i].geom = dCreateCCylinder(space,leg[i].radius,leg[i].length); 
    dGeomSetBody(leg[i].geom,leg[i].body);       
  }

  // hip joint
  joint[HIP] = dJointCreateHinge(world, 0);
  dJointAttach(joint[HIP], torso.body,leg[0].body);
  dJointSetHingeAnchor(joint[HIP], x0, y0, z0 - torso.radius);
  dJointSetHingeAxis(joint[HIP], 1, 0, 0);
  dJointSetHingeParam(joint[HIP], dParamLoStop, -0.25 * M_PI);
  dJointSetHingeParam(joint[HIP], dParamHiStop,  0.25 * M_PI);

  // knee joint
  joint[KNEE] = dJointCreateHinge(world, 0);
  dJointAttach(joint[KNEE],leg[0].body,leg[1].body);
  dJointSetHingeAnchor(joint[KNEE], x0, y0, z0 - torso.radius - leg[0].length);
  dJointSetHingeAxis(joint[KNEE], 1, 0, 0);
  dJointSetHingeParam(joint[KNEE], dParamLoStop, 0.0 * M_PI);
  dJointSetHingeParam(joint[KNEE], dParamHiStop, 0.0 * M_PI);
}


void  setDrawStuff() {
  fn.version = DS_VERSION;
  fn.start   = &start;
  fn.step    = &simLoop;
  fn.command = NULL;
  fn.stop    = NULL;
  fn.path_to_textures = "../../drawstuff/textures";
}

int main (int argc, char **argv)
{
  setDrawStuff();

  world = dWorldCreate();
  space = dHashSpaceCreate(0);
  contactgroup = dJointGroupCreate(0); 

  dWorldSetGravity(world,0,0,-9.8);

  // Create a ground
  ground = dCreatePlane(space,0,0,1,0);

  // create an object
  createMonoBot();

  dsSimulationLoop (argc,argv,352,288,&fn);

  dWorldDestroy(world);

  return 0;
}
